# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import datetime
import sys
import typing

import dimod
import numpy as np

from dwave.samplers.random.cyrandom import sample


__all__ = ['RandomSampler']


class RandomSampler(dimod.Sampler):
    """A random sampler, useful as a performance baseline and for testing.

    Examples:

        >>> from dwave.samplers import RandomSampler
        >>> sampler = RandomSampler()

        Create a random binary quadratic model.

        >>> import dimod
        >>> bqm = dimod.generators.gnp_random_bqm(100, .5, 'BINARY')

        Get 20 random samples.

        >>> sampleset = sampler.sample(bqm, num_reads=20)

        Get the best 5 sample found in .1 seconds

        >>> sampleset = sampler.sample(bqm, time_limit=.1, max_num_samples=5)

    """

    parameters: typing.Mapping[str, typing.List] = dict(
        num_reads=[],
        time_limit=[],
        max_num_samples=[],
        seed=[],
        )
    """Keyword arguments accepted by the sampling methods.

    Examples:

        >>> from dwave.samplers import RandomSampler
        >>> sampler = RandomSampler()
        >>> sampler.parameters
        {'num_reads': [], 'time_limit': [], 'max_num_samples': [], 'seed': []}

    """

    properties: typing.Mapping[str, typing.Any] = dict(
        )
    """Information about the solver. Empty.

    Examples:

        >>> from dwave.samplers import RandomSampler
        >>> sampler = RandomSampler()
        >>> sampler.properties
        {}

    """

    def sample(self,
               bqm: dimod.BinaryQuadraticModel,
               *,
               num_reads: typing.Optional[int] = None,
               time_limit: typing.Optional[typing.Union[float, datetime.timedelta]] = None,
               max_num_samples: int = 1000,
               seed: typing.Union[None, int, np.random.Generator] = None,
               **kwargs,
               ) -> dimod.SampleSet:
        """Return random samples for a binary quadratic model.

        Args:
            bqm: Binary quadratic model to be sampled from.

            num_reads:
                The maximum number of random samples to be drawn.
                If neither ``num_reads`` nor ``time_limit`` are provided,
                ``num_reads`` is set to 1.
                If ``time_limit`` is provided, there is no limit imposed on
                the number of reads.

            time_limit:
                The maximum run time in seconds.
                Only the time to generate the samples, calculate the energies,
                and maintain the population is counted.

            max_num_samples:
                The maximum number of samples returned by the sampler.
                This limits the memory usage for small problems are large
                ``time_limits``.
                Ignored when ``num_reads`` is set.

            seed:
                Seed for the random number generator.
                Passed to :func:`numpy.random.default_rng()`.

        Returns:
            A sample set.
            Some additional information is provided in the
            :attr:`~dimod.SampleSet.info` dictionary:

                * **num_reads**: The total number of samples generated.

        """

        # we could count this towards preprocesing time but IMO it's fine to
        # skip for simplicity.
        self.remove_unknown_kwargs(**kwargs)

        # default case
        if num_reads is None and time_limit is None:
            num_reads = 1
            time_limit = float('inf')

        if num_reads is None:
            # we know that time_limit was set
            num_reads = sys.maxsize
        elif num_reads <= 0:
            raise ValueError("if given, num_reads must be a positive integer")
        else:
            # it was given, so max_num_samples is ignored
            max_num_samples = num_reads

        if time_limit is None:
            # num_reads is specified
            time_limit = float('inf')
        elif isinstance(time_limit, datetime.timedelta):
            time_limit = time_limit.total_seconds()
        elif time_limit <= 0:
            raise ValueError("if given, time_limit must be positive")

        if max_num_samples <= 0:
            raise ValueError("max_num_samples must be a positive integer")

        return sample(bqm,
                      num_reads=num_reads,
                      time_limit=time_limit,
                      max_num_samples=max_num_samples,
                      seed=seed,
                      )



# Copyright 2022 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS F ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import networkx as nx
from dimod import BinaryQuadraticModel


def bqm_to_multigraph(bqm: BinaryQuadraticModel):
    """
    Args:
        bqm: Binary quadratic model to be sampled.
    Returns:
        G - the multi-graph representing the provided BQM, where there is a
            weighted-edge ``uv`` for each quadratic-term, where the weight is
            a linear-function of its bias
        offset - the sum over the biases in the BQM plus its spin-offset
    """
    if any(bqm.spin.linear.values()):
        raise NotImplementedError("not yet implemented for non-zero linear biases")

    offset = bqm.spin.offset
    G = nx.MultiGraph()
    for (u, v), bias in bqm.spin.quadratic.items():
        G.add_edge(u, v, weight=-2 * bias)
        offset += bias

    return G, offset



# Copyright 2019 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcpp cimport bool
from libcpp.vector cimport vector

cimport numpy as np

cdef extern from "descent.h":

    unsigned int steepest_gradient_descent(
        np.int8_t* states,
        double* energies,
        unsigned* num_steps,
        const int num_samples,
        const vector[double]& linear_biases,
        const vector[int]& coupler_starts,
        const vector[int]& coupler_ends,
        const vector[double]& coupler_weights,
        bool large_sparse_opt
    ) nogil



# distutils: language = c++
# cython: language_level = 3

from cython.operator cimport preincrement as inc, dereference as deref

from libcpp.vector cimport vector
from libcpp.unordered_map cimport unordered_map
from libcpp.unordered_set cimport unordered_set

import dimod
cimport dimod


__all__ = ['elimination_order_width', 'min_fill_heuristic']


ctypedef unordered_map[Py_ssize_t, unordered_set[Py_ssize_t]] adj_t

cdef adj_t _cybqm_to_adj(dimod.cyBQM_float64 cybqm):
    cdef adj_t adj

    # add the nodes, even if they don't have any edges
    cdef Py_ssize_t vi
    for vi in range(cybqm.num_variables()):
        adj[vi]

    it = cybqm.data().cbegin_quadratic()
    while it != cybqm.data().cend_quadratic():
        adj[deref(it).u].insert(deref(it).v)
        adj[deref(it).v].insert(deref(it).u)
        inc(it)

    return adj


cdef void _elim_adj(adj_t& adj, Py_ssize_t vi) except +:
    """Remove vi from adj and make its neighborhood a clique."""

    # make the neighborhood of vi a clique
    uit = adj[vi].begin()
    while uit != adj[vi].end():
        vit = uit
        inc(vit)  # no self-loops
        while vit != adj[vi].end():
            adj[deref(vit)].insert(deref(uit))
            adj[deref(uit)].insert(deref(vit))
            inc(vit)

        # remove vi from its neighbors
        adj[deref(uit)].erase(vi)

        inc(uit)

    # finally remove vi
    adj.erase(vi)


def elimination_order_width(bqm, order):
    """Calculates the width of the tree decomposition induced by a
    variable elimination order.

    order must contain exactly the variables of the bqm
    """

    if len(bqm) != len(order):
        raise ValueError("bqm and order must have the name variables")

    cdef dimod.cyBQM_float64 cybqm = dimod.as_bqm(bqm, dtype=float).data

    cdef adj_t adj = _cybqm_to_adj(cybqm)

    # if there is at least one node then the treewidth is at least 1
    cdef Py_ssize_t treewidth = cybqm.num_variables() > 0

    cdef Py_ssize_t vi
    for v in order:
        vi = cybqm.variables.index(v)

        if adj[vi].size() > treewidth:
            treewidth = adj[vi].size()

        _elim_adj(adj, vi)

    return treewidth


# dev note: adj is actually a const reference, but Cython does not like
# that (fixed in Cython 3.0)
cdef Py_ssize_t _min_num_edges(adj_t& adj):
    """Get the node that would need to add the fewest edges when eliminated.

    Only defined for len(adj) > 0.
    """
    # The goal is to go through each node in adj, and determine how many
    # edges we would need to add in order to eliminate it.

    # C++ lambdas don't work so well in Cython so we do 'min' the hard way...

    cdef Py_ssize_t min_num_edges = adj.size() * adj.size()
    cdef Py_ssize_t min_node  # our return value


    cdef Py_ssize_t num_edges
    cdef Py_ssize_t vi

    it = adj.begin()
    while it != adj.end():
        vi = deref(it).first

        # for all pairs of nodes in the neighborhood of vi, count the missing edges
        num_edges = 0
        uit = deref(it).second.begin()
        while uit != deref(it).second.end():
            vit = uit
            while vit != deref(it).second.end():
                if not adj[deref(vit)].count(deref(uit)):
                    num_edges += 1
                inc(vit)
            inc(uit)

        if num_edges < min_num_edges:
            min_num_edges = num_edges
            min_node = vi

        inc(it)

    return min_node


def min_fill_heuristic(bqm):
    """Compute an upper bound on the treewidth of the given bqm based on
    the min-fill heuristic for the elimination ordering.

    Args:
        bqm: a binary quadratic model

    Returns:
        A 2-tuple containing the bound on the treewidth and the elimination 
        order.

    """
    cdef dimod.cyBQM_float64 cybqm = dimod.as_bqm(bqm, dtype=float).data

    cdef adj_t adj = _cybqm_to_adj(cybqm)

    cdef vector[Py_ssize_t] order
    order.reserve(cybqm.num_variables())

    # if there is at least one node then the treewidth is at least 1
    cdef Py_ssize_t upper_bound = cybqm.num_variables() > 0

    cdef Py_ssize_t vi
    while adj.size():
        vi = _min_num_edges(adj)

        if adj[vi].size() > upper_bound:
            upper_bound = adj[vi].size()

        # remove vi from adj
        _elim_adj(adj, vi)

        order.push_back(vi)

    cdef Py_ssize_t i
    variables = []
    for i in range(order.size()):
        variables.append(cybqm.variables.at(order[i]))

    return upper_bound, variables



# distutils: language = c++
# cython: language_level = 3
# distutils: include_dirs = dwave/samplers/tree/src/include/

# Copyright 2019 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dimod
import numpy as np

cimport numpy as np
from cython.operator cimport dereference as deref
from dimod cimport cyBQM_float64
from dimod.libcpp cimport BinaryQuadraticModel as cppBinaryQuadraticModel
from dimod.binary.binary_quadratic_model import BinaryQuadraticModel

from dwave.samplers.tree.orang cimport energies_type, samples_type, PyArray_ENABLEFLAGS

cdef extern from "src/include/solve.hpp":
    void solveBQM[V, B](cppBinaryQuadraticModel[B, V]& refBQM,
                        int* var_order,
                        double beta,
                        int low,
                        double max_complexity,
                        int max_solutions,
                        double** energies_data, int* energies_len,
                        int** sols_data, int* sols_rows, int* sols_cols) except +

samples_dtype = np.intc  # needs to be consistent with samples_type
energies_dtype = np.double  # needs to be consistent with energies_type 

def solve_bqm_wrapper(bqm: BinaryQuadraticModel,
                      order: list,
                      max_complexity: int,
                      max_solutions: int = 1):
    """Cython wrapper for :func:`solveBQM`.

    Args:
        bqm:
            Binary quadratic model to sample from. Variables must be linearly indexed.

        order:
            List of variables representing the variable elimination order.

        max_complexity:
            Upper bound on algorithm's complexity.

        max_solutions:
            Maximum number of solutions to find.

    Returns:
        The samples and marginals.
    """

    if not bqm.num_variables:
        raise ValueError("bqm must have at least one variable.")

    if len(order) != bqm.num_variables or set(order) != bqm.variables:
        raise ValueError("order must contain the variables in bqm.")

    cdef cyBQM_float64 cybqm = bqm.data
    cdef double beta = -1  # solving
    cdef double _max_complexity = max_complexity
    cdef int low = -1 if bqm.vartype is dimod.SPIN else 0

    cdef int[:] elimination_order = np.asarray(order, dtype=np.intc)
    cdef int* elimination_order_ptr = &elimination_order[0]

    # Pass in a pointer so that solveBQM can fill it in. Note that this is
    # a design choice inherited from orang's c++ implementation. In the future
    # we may want to change it.
    cdef int num_energies, srows, scols
    cdef energies_type* energies_pointer
    cdef samples_type* samples_pointer

    solveBQM(deref(cybqm.cppbqm),
             elimination_order_ptr,
             beta,
             low,
             _max_complexity,
             max_solutions,
             &energies_pointer, &num_energies,
             &samples_pointer, &srows, &scols
            )

    # create a numpy array without making a copy then tell numpy it needs to
    # free the memory
    samples = np.asarray(<samples_type[:srows, :scols]> samples_pointer)
    PyArray_ENABLEFLAGS(samples, np.NPY_OWNDATA)

    # convert the samples to spin if necessary
    cdef size_t i, j
    if low == -1:
        for i in range(srows):
            for j in range(scols):
                if samples[i, j] == 0:
                    samples[i, j] = -1

    energies = np.asarray(<energies_type[:num_energies]> energies_pointer)

    PyArray_ENABLEFLAGS(energies, np.NPY_OWNDATA)

    energies += bqm.offset

    return samples, energies



# distutils: language = c++
# distutils: include_dirs = dwave/samplers/sa/src/
# distutils: sources = dwave/samplers/sa/src/cpu_sa.cpp
# cython: language_level = 3

# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from libcpp cimport bool
from libcpp.vector cimport vector

import numpy as np
cimport numpy as np


cdef extern from "cpu_sa.h":
    ctypedef bool (*callback)(void *function)
    ctypedef enum Proposal:
        Gibbs, Metropolis
    ctypedef enum VariableOrder:
        Sequential, Random
    int general_simulated_annealing(
            np.int8_t* samples,
            double* energies,
            const int num_samples,
            const vector[double] & h,
            const vector[int] & coupler_starts,
            const vector[int] & coupler_ends,
            const vector[double] & coupler_weights,
            const int sweeps_per_beta,
            const vector[double] & beta_schedule,
            const unsigned long long seed,
            const VariableOrder varorder,
            const Proposal proposal_acceptance_criteria,
            callback interrupt_callback,
            void *interrupt_function) nogil


def simulated_annealing(num_samples, h, coupler_starts, coupler_ends,
                        coupler_weights, sweeps_per_beta, beta_schedule, seed,
                        np.ndarray[np.int8_t, ndim=2, mode="c"] states_numpy,
                        randomize_order=False,
                        proposal_acceptance_criteria='Metropolis',
                        interrupt_function=None):
    """Wraps `general_simulated_annealing` from `cpu_sa.cpp`. Accepts
    an Ising problem defined on a general graph and returns samples
    using simulated annealing.

    Parameters
    ----------
    num_samples : int
        Number of samples to get from the sampler.

    h : list(float)
        The h or field values for the problem.

    coupler_starts : list(int)
        A list of the start variable of each coupler. For a problem
        with the couplers (0, 1), (1, 2), and (3, 1), `coupler_starts`
        should be [0, 1, 3].

    coupler_ends : list(int)
        A list of the end variable of each coupler. For a problem
        with the couplers (0, 1), (1, 2), and (3, 1), `coupler_ends`
        should be [1, 2, 1].

    coupler_weights : list(float)
        A list of the J values or weight on each coupler, in the same
        order as `coupler_starts` and `coupler_ends`.

    sweeps_per_beta : int
        The number of sweeps to perform at each beta value provided in
        `beta_schedule`. The total number of sweeps per sample is
        sweeps_per_beta * len(beta_schedule).

    beta_schedule : list(float)
        A list of the different beta values to run sweeps at.

    seed : 64 bit int > 0
        The seed to use for the PRNG. Must be a positive integer
        greater than 0. If the same seed is used and the rest of the
        parameters are the same, the returned samples will be
        identical.

    states_numpy : np.ndarray[int8_t, ndim=2, mode="c"], values in (-1, 1)
        The initial seeded states of the simulated annealing runs. Should be of
        a contiguous numpy.ndarray of shape (num_samples, num_variables).

    interrupt_function: function
        Should accept no arguments and return a bool. The function is
        called between samples and if it returns True, simulated annealing
        will return early with the samples it already has.

    randomize_order: bool
        When True, each spin update selects a variable uniformly at random.
        When False, updates proceed sequentially through the labeled variables 
        on each sweep so that all variables are updated once per sweep.
        The random method is ergodic and obeys detailed balance at all temperatures.
        Symmetries of the Boltzmann distribution(s) are not broken by the
        update order.
        The sequential method:
            - when combined with ``proposal_acceptance_criteria=Metropolis`` can be
            non-ergodic in the limits of zero or infinite temperature,
            and converge slowly near these limits.
            - can introduce a dynamical bias as a function of variable
            labeling convention.
            - has faster per spin update than the random method.

    proposal_acceptance_criteria: str
        When `Gibbs`, each spin flip proposal is accepted according to the
        Gibbs criteria.
        When `Metropolis`, each spin flip proposal is accepted according to the
        Metropolis-Hastings criteria.

    Returns
    -------
    samples : numpy.ndarray
        A 2D numpy array where each row is a sample.

    energies: np.ndarray
        The energies.

    """
    num_vars = len(h)

    # in the case that we either need no samples or there are no variables,
    # we can safely return an empty array (and set energies to 0)
    if num_samples*num_vars == 0:
        annealed_states = np.empty((num_samples, num_vars), dtype=np.int8)
        return annealed_states, np.zeros(num_samples, dtype=np.double)

    # allocate ndarray for energies
    energies_numpy = np.empty(num_samples, dtype=np.float64)
    cdef double[:] energies = energies_numpy

    # explicitly convert all Python types to C while we have the GIL
    cdef np.int8_t* _states = &states_numpy[0, 0]
    cdef double* _energies = &energies[0]
    cdef int _num_samples = num_samples
    cdef vector[double] _h = h
    cdef vector[int] _coupler_starts = coupler_starts
    cdef vector[int] _coupler_ends = coupler_ends
    cdef vector[double] _coupler_weights = coupler_weights
    cdef int _sweeps_per_beta = sweeps_per_beta
    cdef vector[double] _beta_schedule = beta_schedule
    cdef unsigned long long _seed = seed
    cdef VariableOrder _varorder
    if randomize_order:
        _varorder = Random
    else:
        _varorder = Sequential
    cdef Proposal _proposal_acceptance_criteria
    if proposal_acceptance_criteria.lower() == 'gibbs':
        _proposal_acceptance_criteria = Gibbs
    elif proposal_acceptance_criteria.lower() == 'metropolis':
        _proposal_acceptance_criteria = Metropolis
    else:
        raise ValueError(f'Unknown proposal_acceptance_criteria: {proposal_acceptance_criteria}')
    cdef void* _interrupt_function
    if interrupt_function is None:
        _interrupt_function = NULL
    else:
        _interrupt_function = <void *>interrupt_function


    with nogil:
        num = general_simulated_annealing(_states,
                                          _energies,
                                          _num_samples,
                                          _h,
                                          _coupler_starts,
                                          _coupler_ends,
                                          _coupler_weights,
                                          _sweeps_per_beta,
                                          _beta_schedule,
                                          _seed,
                                          _varorder,
                                          _proposal_acceptance_criteria,
                                          interrupt_callback,
                                          _interrupt_function)

    # discard the noise if we were interrupted
    return states_numpy[:num], energies_numpy[:num]


cdef bool interrupt_callback(void * const interrupt_function) with gil:
    try:
        return (<object>interrupt_function)()
    except Exception:
        # if an exception occurs, treat as an interrupt
        return True



# Copyright 2022 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS F ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import typing

import dimod
import networkx as nx

from dwave.samplers.planar.planar import rotation_from_coordinates, plane_triangulate, expanded_dual
from dwave.samplers.planar.util import bqm_to_multigraph

__all__ = ["PlanarGraphSolver"]


class PlanarGraphSolver(dimod.Sampler, dimod.Initialized):
    """An exact solver for planar Ising problems with no linear biases."""
    parameters: typing.Dict[str, typing.Sequence[str]] = dict(pos=tuple())
    """Keyword arguments accepted by the sampling methods."""
    properties: typing.Dict[str, typing.Any] = dict()
    """Values for parameters accepted by the sampling methods."""

    def __init__(self):
        self.parameters = copy.deepcopy(self.parameters)
        self.properties = copy.deepcopy(self.properties)

    def sample(self,
               bqm: dimod.BinaryQuadraticModel,
               pos: typing.Optional[typing.Mapping[dimod.typing.Variable, typing.Tuple[float, float]]] = None,
               **kwargs) -> dimod.SampleSet:
        """Find a ground state of a planar Ising problem without linear biases.

        Args:
            bqm: Binary quadratic model representing a planar Ising model with no linear biases.
            pos: Position for each node as a dict of form ``{node: (x-coordinate, y-coordinate),...}``.

        Examples:
            >>> import dimod
            >>> from dwave.samplers.planar import PlanarGraphSolver
            >>> bqm = dimod.BinaryQuadraticModel.empty(dimod.SPIN)
            >>> bqm.add_interaction('a', 'b', +1.0)
            >>> bqm.add_interaction('b', 'c', +1.0)
            >>> bqm.add_interaction('c', 'a', +1.0)
            >>> pos = {'a': (0, 0), 'b': (1, 0), 'c': (0, 1)}
            >>> sample = PlanarGraphSolver().sample(bqm, pos)
            >>> sample.first
            Sample(sample={'a': 1, 'b': -1, 'c': -1}, energy=-1.0, num_occurrences=1)

        """

        if len(bqm) < 3:
            raise ValueError("The provided BQM must have at least three variables")

        G, off = bqm_to_multigraph(bqm)

        if pos is None:
            pos = _determine_pos(G)

        # apply the rotation system
        r = rotation_from_coordinates(G, pos)
        nx.set_node_attributes(G, name='rotation', values=r)

        # triangulation
        plane_triangulate(G)

        # create an edge indexing scheme
        indices = {edge: idx for idx, edge in enumerate(G.edges(keys=True))}
        nx.set_edge_attributes(G, name='index', values=indices)

        dual = expanded_dual(G)
        matching = nx.max_weight_matching(dual, maxcardinality=True, weight='weight')

        assert nx.is_perfect_matching(dual, matching)

        cut = _dual_matching_to_cut(G, matching)
        state = _cut_to_state(G, cut)

        if bqm.vartype is not dimod.BINARY:
            state = {v: 2 * b - 1 for v, b in state.items()}

        return dimod.SampleSet.from_samples_bqm(state, bqm)


def _determine_pos(G: nx.MultiGraph) -> dict:
    is_planar, P = nx.check_planarity(G)
    if not is_planar:
        raise ValueError("The provided BQM does not yield a planar embedding")

    return nx.planar_layout(P)


def _dual_matching_to_cut(G: nx.MultiGraph, matching: set) -> set:
    """
    Then for each edge u,v in G, the expanded dual has two nodes (u,v), (v,u) with
    an edge between them. The matching is defined on that graph. There are
    additional edges, but we don't care about those
    """
    cut = set(G.edges)

    # need to get the cut from the matching
    for u, v in matching:
        if u[0] == v[1] and u[1] == v[0] and u[2] == v[2]:
            cut.discard(u)
            cut.discard(v)

    return cut


def _cut_to_state(G: nx.MultiGraph, cut: set, node=None, val=0) -> dict:
    if node is None:
        node = next(iter(G))  # get any node and assign it to val

    state = {}

    def _cut_to_state0(v, s):
        if v in state:
            assert state[v] == s, "Inconsistent state"
        else:
            state[v] = s
            for u in G[v]:
                for key in G[u][v]:
                    if (u, v, key) in cut or (v, u, key) in cut:
                        _cut_to_state0(u, 1 - s)
                    else:
                        _cut_to_state0(u, s)

    _cut_to_state0(node, val)

    return state



# distutils: language = c++
# cython: language_level = 3

# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

cimport cython

from cpython.pycapsule cimport PyCapsule_IsValid, PyCapsule_GetPointer
from libcpp.algorithm cimport make_heap, push_heap, pop_heap
from libcpp.utility cimport pair
from libcpp.vector cimport vector

import dimod
cimport dimod
import numpy as np
cimport numpy as np
cimport numpy.random

# chrono is not included in Cython's libcpp. So we do it more manually
cdef extern from *:
    """
    #include <chrono>

    double realtime_clock() {
        auto t = std::chrono::high_resolution_clock::now();
        return std::chrono::duration<double>(t.time_since_epoch()).count();
    }
    """
    double realtime_clock()


# it would be nicer to use a struct, but OSX throws segfaults when sorting
# Cython-created structs. We should retest that when we switch to Cython 3.
ctypedef pair[np.float64_t, vector[np.int8_t]] state_t


cdef state_t get_sample(dimod.cyBQM_float64 cybqm,
                        numpy.random.bitgen_t* bitgen,
                        bint is_spin = False,
                        ):
    # developer note: there is a bunch of potential optimization here
    cdef state_t state

    # generate the sample
    state.second.reserve(cybqm.num_variables())
    cdef Py_ssize_t i
    for i in range(cybqm.num_variables()):
        state.second.push_back(bitgen.next_uint32(bitgen.state) % 2)

    if is_spin:
        # go back through and convert to spin
        for i in range(state.second.size()):
            state.second[i] = 2 * state.second[i] - 1

    state.first = cybqm.data().energy(state.second.begin())

    return state


@cython.boundscheck(False)
@cython.wraparound(False)
def sample(object bqm,
           Py_ssize_t num_reads,
           np.float64_t time_limit,
           Py_ssize_t max_num_samples,
           object seed):

    cdef Py_ssize_t i, j  # counters for use later

    # Get Cython access to the BQM. We could template to avoid the copy,
    # but honestly everyone just uses float64 anyway so...
    cdef dimod.cyBQM_float64 cybqm = dimod.as_bqm(bqm, dtype=float).data
    cdef bint is_spin = bqm.vartype is dimod.SPIN

    if num_reads <= 0:
        raise ValueError("num_reads must be positive")
    if time_limit <= 0:
        raise ValueError("time_limit must be positive")
    if max_num_samples <= 0:
        raise ValueError("max_num_samples must be positive")

    # Get Cython access to the rng
    rng = np.random.default_rng(seed)
    cdef numpy.random.bitgen_t *bitgen
    cdef const char *capsule_name = "BitGenerator"
    capsule = rng.bit_generator.capsule
    if not PyCapsule_IsValid(capsule, capsule_name):
        raise ValueError("Invalid pointer to anon_func_state")
    bitgen = <numpy.random.bitgen_t *> PyCapsule_GetPointer(capsule, capsule_name)

    # ok, time to start sampling!
    cdef double sampling_start_time = realtime_clock()
    cdef double sampling_stop_time = sampling_start_time + time_limit

    # fill out the population
    cdef vector[state_t] samples
    for i in range(max_num_samples):
        samples.push_back(get_sample(cybqm, bitgen, is_spin))

        if samples.size() >= num_reads:
            break
        if realtime_clock() >= sampling_stop_time:
            break

    cdef Py_ssize_t num_drawn = samples.size()

    # determine if we need to keep going
    if samples.size() < num_reads and realtime_clock() < sampling_stop_time:
        # at this point, we want to stop growing our samples vector, so
        # we turn it into a heap
        make_heap(samples.begin(), samples.end())

        # we're never going to change size, so we stop testing the num_reads
        # condition. It was up to the caller to ensure that max_num_samples >= num_reads
        # if they want to terminate on num_reads
        while realtime_clock() < sampling_stop_time:
            samples.push_back(get_sample(cybqm, bitgen, is_spin))
            push_heap(samples.begin(), samples.end())
            pop_heap(samples.begin(), samples.end())
            samples.pop_back()

            num_drawn += 1

    # sampling done!
    # time to construct the return objects

    record = np.rec.array(
        np.empty(samples.size(),
                 dtype=[('sample', np.int8, (bqm.num_variables,)),
                        ('energy', float),
                        ('num_occurrences', int)]))

    record['num_occurrences'][:] = 1

    cdef np.float64_t[:] energies_view = record['energy']
    for i in range(samples.size()):
        energies_view[i] = samples[i].first

    cdef np.int8_t[:, :] sample_view = record['sample']
    for i in range(samples.size()):
        for j in range(cybqm.num_variables()):
            sample_view[i, j] = samples[i].second[j]

    sampleset = dimod.SampleSet(record, bqm.variables, info=dict(), vartype=bqm.vartype)

    sampleset.info.update(
        num_reads=num_drawn,
        # todo: timing data
        )

    return sampleset



# Copyright 2018 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS F ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A dimod :term:`sampler` that uses the MST2 multistart tabu search algorithm."""

import numpy as np
import dimod

from typing import Optional

from dimod.core.initialized import InitialStateGenerator
from dwave.samplers.tabu import TabuSearch

__all__ = ["TabuSampler"]


class TabuSampler(dimod.Sampler, dimod.Initialized):
    """A tabu-search sampler.

    `Tabu search <https://en.wikipedia.org/wiki/Tabu_search>`_ is a heuristic that
    employs local search and can escape local minima by maintaining a "tabu list" of
    recently explored states that it does not revisit. This sampler implements the
    `MST2 multistart tabu search algorithm.
    <https://link.springer.com/article/10.1023/B:ANOR.0000039522.58036.68>`_
    for quadratic unconstrained binary optimization (QUBO) problems.

    Examples:
        This example solves a two-variable Ising model.

        >>> from dwave.samplers import TabuSampler
        >>> samples = TabuSampler().sample_ising({'a': -0.5, 'b': 1.0}, {'ab': -1})
        >>> list(samples.data()) # doctest: +SKIP
        [Sample(sample={'a': -1, 'b': -1}, energy=-1.5, num_occurrences=1)]
        >>> samples.first.energy
        -1.5

    """

    properties = None
    parameters = None

    def __init__(self):
        self.parameters = {
            'initial_states': [],
            'initial_states_generator': [],
            'num_reads': [],
            'seed': [],
            'tenure': [],
            'timeout': [],
            'num_restarts': [],
            'energy_threshold': [],
            'coefficient_z_first': [],
            'coefficient_z_restart': [],
            'lower_bound_z': [],
        }
        self.properties = {}

    def sample(self, bqm: dimod.BinaryQuadraticModel,
               initial_states: Optional[dimod.typing.SamplesLike] = None,
               initial_states_generator: InitialStateGenerator = 'random',
               num_reads: Optional[int] = None,
               seed: Optional[int] = None,
               tenure: Optional[int] = None,
               timeout: int = 20,
               num_restarts: int = 1000000,
               energy_threshold: Optional[float] = None,
               coefficient_z_first: Optional[int] = None,
               coefficient_z_restart: Optional[int] = None,
               lower_bound_z: Optional[int] = None,
               **kwargs) -> dimod.SampleSet:
        """Run a multistart tabu search on a given binary quadratic model.

        Args:
            bqm: Binary quadratic model (BQM) to be sampled.

            initial_states:
                One or more samples, each defining an initial state for all the
                problem variables. Initial states are given one per read, but
                if fewer than ``num_reads`` initial states are defined, additional
                values are generated as specified by ``initial_states_generator``.

            initial_states_generator:
                Defines the expansion of ``initial_states`` if fewer than
                ``num_reads`` are specified:

                * "none":
                    If the number of initial states specified is smaller than
                    ``num_reads``, raises ``ValueError``.

                * "tile":
                    Reuses the specified initial states if fewer than ``num_reads``
                    or truncates if greater.

                * "random":
                    Expands the specified initial states with randomly generated
                    states if fewer than ``num_reads`` or truncates if greater.

            num_reads:
                Number of reads. Each read is generated by one run of the tabu
                algorithm. If ``num_reads`` is not explicitly given, it is selected
                to match the number of initial states given. If initial states
                are not provided, only one read is performed.

            seed:
                32-bit unsigned integer seed to use for the PRNG. If the ``timeout``
                parameter is not None, results from the same seed may not be
                identical between runs due to finite clock resolution.

            tenure:
                Tabu tenure, which is the length of the tabu list, or number of
                recently explored solutions kept in memory.
                Default is a quarter of the number of problem variables up to
                a maximum value of 20.

            timeout:
                Maximum running time per read in milliseconds.

            num_restarts:
                Maximum number of tabu search restarts per read. Setting this value
                to zero results in a simple tabu search. 

            energy_threshold:
                Terminate when an energy lower than or equal to ``energy_threshold`` is found.

            coefficient_z_first:
                :code:`max(bqm.num_variables*coefficient_z_first, lower_bound_z)`
                bounds the number of variable updates considered in
                the first simple tabu search (STS). Variable updates arising from 
                the STS greedy-descent subroutine, invoked upon discovery of new 
                global optima, are excluded from the count. The 
                coefficient defaults to 10000 for small problems (up to 500 
                variables) and 25000 for larger problems.
                
            coefficient_z_restart:
                Controls the number of variable updates on restarted simple tabu
                search stages, matching the description for ``coefficient_z_first``. 
                The coefficient defaults to :code:`coefficient_z_first/4`
                
            lower_bound_z:
                Sets a minimum number of variable updates on all simple tabu
                searches, see ``coefficient_z_first``. The bound defaults to 
                500000. 

        Examples:
            This example samples a simple two-variable Ising model.

            >>> import dimod
            >>> bqm = dimod.BQM.from_ising({}, {'ab': 1})

            >>> from dwave.samplers import TabuSampler
            >>> sampler = TabuSampler()

            >>> samples = sampler.sample(bqm)
            >>> samples.first.energy
            -1.0
        """

        if not bqm:
            return dimod.SampleSet.from_samples([], energy=0, vartype=bqm.vartype)

        if tenure is None:
            tenure = min(20, len(bqm) // 4)
        elif not isinstance(tenure, int):
            raise TypeError("'tenure' should be an integer in range [0, num_vars - 1]")
        elif not 0 <= tenure < len(bqm):
            raise ValueError("'tenure' should be an integer in range [0, num_vars - 1]")

        # Get initial_states in binary form
        parsed = self.parse_initial_states(bqm.binary,
                                           initial_states=initial_states,
                                           initial_states_generator=initial_states_generator,
                                           num_reads=num_reads,
                                           seed=seed)

        parsed_initial_states = np.ascontiguousarray(parsed.initial_states.record.sample)

        qubo, varorder = self._bqm_to_tabu_qubo(bqm.binary)

        if timeout is None:
            timeout = -1    # Using negative timeout to mean ignore timeout parameter

        # run Tabu search
        samples = np.empty((parsed.num_reads, len(bqm)), dtype=np.int8)

        rng = np.random.default_rng(seed)

        restarts = []
        for ni, initial_state in enumerate(parsed_initial_states):
            seed_per_read = rng.integers(2**32, dtype=np.uint32)
            r = TabuSearch(qubo, initial_state, tenure, timeout, num_restarts, seed_per_read, energy_threshold, coefficient_z_first, coefficient_z_restart, lower_bound_z)
            samples[ni, :] = r.bestSolution()
            restarts.append(r.numRestarts())

        # we received samples in binary form, so convert if needed
        if bqm.vartype is dimod.SPIN:
            samples *= 2
            samples -= 1
        elif bqm.vartype is not dimod.BINARY:
            # sanity check
            raise ValueError("unknown vartype")

        return dimod.SampleSet.from_samples_bqm((samples, varorder), bqm=bqm, num_restarts=restarts)

    @staticmethod
    def _bqm_to_tabu_qubo(bqm):
        # construct dense matrix representation
        ldata, (irow, icol, qdata), offset, varorder = bqm.binary.to_numpy_vectors(return_labels=True)
        ud = np.zeros((len(bqm), len(bqm)), dtype=np.double)
        ud[np.diag_indices(len(bqm), 2)] = ldata
        ud[irow, icol] = qdata

        # Note: normally, conversion would be: `ud + ud.T - np.diag(np.diag(ud))`,
        # but the Tabu solver we're using requires slightly different qubo matrix.
        ud *= .5
        symm = ud + ud.T
        return symm, varorder



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import math

from numbers import Integral
from numpy.random import randint
from collections import defaultdict
from typing import List, Sequence, Tuple, Optional, Union
from time import perf_counter_ns
try:
    from typing import Literal
    BetaScheduleType = Literal['linear', 'geometric', 'custom']
except ImportError:
    BetaScheduleType = str

    
from dimod.core.initialized import InitialStateGenerator

import dimod
import numpy as np

from dwave.samplers.sa.simulated_annealing import simulated_annealing

import warnings

__all__ = ["Neal", "SimulatedAnnealingSampler"]


class SimulatedAnnealingSampler(dimod.Sampler, dimod.Initialized):
    """Simulated annealing sampler for binary quadratic models.

    `Simulated annealing <https://en.wikipedia.org/wiki/Simulated_annealing>`_
    can be used for heuristic optimization or approximate Boltzmann sampling. This
    implementation approaches the equilibrium distribution by performing updates
    at a sequence of decreasing temperatures, terminating at the target
    :math:`\\beta`.\ [#]_ By default each spin is updated once in a fixed order per point
    per :math:`\\beta` according to a Metropolis-Hastings update. When :math:`\\beta`
    is large the target distribution concentrates, at equilibrium, over ground
    states of the model. Samples are guaranteed to match the equilibrium for
    long, smooth :math:`\\beta` schedules.

    .. [#] :math:`\\beta` represents the inverse temperature, :math:`1/(k_B T)`,
       of a `Boltzmann distribution <https://en.wikipedia.org/wiki/Boltzmann_distribution>`_
       where :math:`T` is the thermodynamic temperature in kelvin and :math:`k_B`
       is Boltzmann's constant.

    For more information, see Kirkpatrick, S.; Gelatt Jr, C. D.; Vecchi, M. P.
    (1983). "Optimization by Simulated Annealing". Science. 220 (4598): 671–680

    Also aliased as :class:`.Neal`.

    Examples:
        This example solves a simple Ising problem.

        >>> from dwave.samplers import SimulatedAnnealingSampler
        >>> sampler = SimulatedAnnealingSampler()
        >>> h = {'a': 0.0, 'b': 0.0, 'c': 0.0}
        >>> J = {('a', 'b'): 1.0, ('b', 'c'): 1.0, ('a', 'c'): 1.0}
        >>> sampleset = sampler.sample_ising(h, J, num_reads=10)
        >>> print(sampleset.first.energy)
        -1.0

    """

    parameters = None
    """Keyword arguments accepted by the sampling methods.

    See :meth:`.SimulatedAnnealingSampler.sample` for a description of the
    parameters.

    Examples:
        This example looks at a sampler's parameters and some of their values.

        >>> from dwave.samplers import SimulatedAnnealingSampler
        >>> sampler = SimulatedAnnealingSampler()
        >>> for kwarg in sorted(sampler.parameters):
        ...     print(kwarg)
        beta_range
        beta_schedule_type
        initial_states
        initial_states_generator
        interrupt_function
        num_reads
        num_sweeps
        num_sweeps_per_beta
        seed
        >>> sampler.parameters['beta_range']
        []
        >>> sampler.parameters['beta_schedule_type']
        ['beta_schedule_options']

    """

    properties = None
    """A dict containing any additional information about the sampler.

    Examples:
        This example looks at the supported values of a sampler property.

        >>> from dwave.samplers import SimulatedAnnealingSampler
        >>> sampler = SimulatedAnnealingSampler()
        >>> sampler.properties['beta_schedule_options']
        ('linear', 'geometric', 'custom')

    """

    def __init__(self):
        # create a local copy in case folks for some reason want to modify them
        self.parameters = {'beta_range': [],
                           'num_reads': [],
                           'num_sweeps': [],
                           'num_sweeps_per_beta': [],
                           'beta_schedule_type': ['beta_schedule_options'],
                           'seed': [],
                           'interrupt_function': [],
                           'initial_states': [],
                           'initial_states_generator': [],
                           }
        self.properties = {'beta_schedule_options': ('linear', 'geometric',
                                                     'custom')}

    def sample(self, bqm: dimod.BinaryQuadraticModel,
               beta_range: Optional[Union[List[float], Tuple[float, float]]] = None,
               num_reads: Optional[int] = None,
               num_sweeps: Optional[int] = None,
               num_sweeps_per_beta: int = 1,
               beta_schedule_type: BetaScheduleType = "geometric",
               seed: Optional[int] = None,
               interrupt_function = None,
               beta_schedule: Optional[Union[Sequence[float], np.ndarray]] = None,
               initial_states: Optional[dimod.typing.SamplesLike] = None,
               initial_states_generator: InitialStateGenerator = "random",
               randomize_order: bool = False,
               proposal_acceptance_criteria: str = 'Metropolis',
               **kwargs) -> dimod.SampleSet:
        """Sample from a binary quadratic model.

        Args:
            bqm: Binary quadratic model to be sampled.

            beta_range:
                A 2-tuple or list defining the beginning and end of the
                :math:`\\beta`\ [#]_ schedule. The schedule is
                interpolated within this range according to the value specified
                by ``beta_schedule_type``. Default range is set based on the
                total bias associated with each node.

            num_reads:
                Number of reads. Each read is generated by one run of the
                simulated annealing algorithm. If ``num_reads`` is not explicitly
                given, it is selected to match the number of initial states
                given. If initial states are not provided, only one read is
                performed.

            num_sweeps:
                Number of sweeps used in annealing. If no value is provided
                and ``beta_schedule`` is None, the value defaults to 1000.

            num_sweeps_per_beta (int, optional, default=1)
                Number of sweeps to perform at each :math:`\\beta`. One sweep
                consists of a sequential Metropolis update of all spins.

            beta_schedule_type:
                :math:`\\beta` schedule type, or how the :math:`\\beta` values
                are interpolated  between the given ``beta_range``. Supported
                values are:

                * "linear"

                * "geometric"

                * "custom"

                "custom" is recommended for high-performance applications, which
                typically require optimizing :math:`\\beta` schedules beyond those
                of the "linear" and "geometric" options, with bounds beyond those
                provided by default. ``num_sweeps_per_beta`` and
                ``beta_schedule`` fully specify a custom schedule.

            beta_schedule:
                Sequence of :math:`\\beta` values swept. Format must be compatible
                with ``numpy.array(beta_schedule, dtype=float)``. Values should
                be non-negative.

            seed:
                Seed to use for the PRNG. Specifying a particular seed with a
                constant set of parameters produces identical results. If not
                provided, a random seed is chosen.

            initial_states:
                One or more samples, each defining an initial state for all the
                problem variables. Initial states are given one per read, but
                if fewer than ``num_reads`` initial states are defined,
                additional values are generated as specified by
                ``initial_states_generator``. See :func:`~dimod.as_samples` for a
                description of "samples-like".

            initial_states_generator:
                Defines the expansion of ``initial_states`` if fewer than
                ``num_reads`` are specified:

                * "none":
                    If the number of initial states specified is smaller than
                    ``num_reads``, raises ``ValueError``.

                * "tile":
                    Reuses the specified initial states if fewer than
                    ``num_reads`` or truncates if greater.

                * "random":
                    Expands the specified initial states with randomly generated
                    states if fewer than ``num_reads`` or truncates if greater.

            randomize_order:
                When `True`, each spin update selects a variable uniformly at random.
                This method is ergodic, obeys detailed balance and preserves symmetries 
                of the model.

                When `False`, updates proceed sequentially through the labeled variables 
                on each sweep so that all variables are updated once per sweep. This method:

                * can be non-ergodic in special cases when used with ``proposal_acceptance_critera=="Metropolis"``.

                * can introduce a dynamical bias as a function of variable order.

                * has faster per spin update than the True method.

            proposal_acceptance_criteria:
                When "Gibbs", each spin flip proposal is accepted according
                to the Gibbs criteria.
                When "Metropolis", each spin flip proposal is accepted according
                to the Metropolis-Hastings criteria.

            interrupt_function (function, optional):
                A function called with no parameters between each sample of
                simulated annealing. If the function returns True, simulated
                annealing terminates and returns with all of the samples and
                energies found so far.

        Returns:
            A `dimod.SampleSet` for the binary quadratic model.

            The `info` field of the sample set contains information about the sampling procedure:
            1. the beta range used,
            2. the beta schedule type used, and
            3. timing information (details below).

            Timing information is categorized into three: preprocessing, sampling, and
            postprocessing time. All timings are reported in units of nanoseconds.

            Preprocessing time is the total time spent converting the BQM variable type (if
            required), parsing input arguments, and determining an annealing schedule. Sampling time
            is the total time the algorithm spent in sampling states of the binary quadratic model.
            Postprocessing time is the total time spent reverting variable type and creating a
            `dimod.SampleSet`.

        Examples:
            This example runs simulated annealing on a binary quadratic model
            with various input parameters.

            >>> import dimod
            >>> from dwave.samplers import SimulatedAnnealingSampler
            ...
            >>> sampler = SimulatedAnnealingSampler()
            >>> bqm = dimod.BinaryQuadraticModel({'a': .5, 'b': -.5},
            ...                                  {('a', 'b'): -1}, 0.0,
            ...                                  dimod.SPIN)
            >>> # Run with default parameters
            >>> sampleset = sampler.sample(bqm)
            >>> # Run with specified parameters
            >>> sampleset = sampler.sample(bqm, seed=1234,
            ...                            beta_range=[0.1, 4.2],
            ...                            num_sweeps=20,
            ...                            beta_schedule_type='geometric')
            >>> # Reuse a seed
            >>> a1 = next((sampler.sample(bqm, seed=88)).samples())['a']
            >>> a2 = next((sampler.sample(bqm, seed=88)).samples())['a']
            >>> a1 == a2
            True

            .. [#] :math:`\\beta` represents the inverse temperature, :math:`1/(k_B T)`, of a
               `Boltzmann distribution <https://en.wikipedia.org/wiki/Boltzmann_distribution>`_
               where :math:`T` is the thermodynamic temperature in kelvin and :math:`k_B` is
               Boltzmann's constant.

        """
        timestamp_preprocess = perf_counter_ns()
        # get the original vartype so we can return consistently
        original_vartype = bqm.vartype

        # convert to spin (if needed)
        if bqm.vartype is not dimod.SPIN:
            bqm = bqm.change_vartype(dimod.SPIN, inplace=False)

        # parse_initial_states could handle seed generation, but because we're
        # sharing it with the SA algo, we handle it out here
        if seed is None:
            seed = randint(2**31)
        elif not isinstance(seed, Integral):
            error_msg = ("'seed' should be None or an integer between 0 and 2^32 "
                         "- 1: value = {}".format(seed))
            raise TypeError(error_msg)
        elif not (0 <= seed < 2**31):
            error_msg = ("'seed' should be an integer between 0 and 2^32 - 1: "
                         "value = {}".format(seed))
            raise ValueError(error_msg)

        # parse the inputs
        parsed = self.parse_initial_states(
            bqm,
            num_reads=num_reads,
            initial_states=initial_states,
            initial_states_generator=initial_states_generator,
            seed=seed)

        num_reads = parsed.num_reads

        # read out the initial states and the variable order
        initial_states_array = np.ascontiguousarray(
            parsed.initial_states.record.sample)

        variable_order = parsed.initial_states.variables

        # read out the BQM
        ldata, (irow, icol, qdata), off = bqm.to_numpy_vectors(
            variable_order=variable_order)

        if interrupt_function and not callable(interrupt_function):
            raise TypeError("'interrupt_function' should be a callable")

        if not isinstance(num_sweeps_per_beta, Integral):
            error_msg = "'num_sweeps_per_beta' should be a positive integer: value = {}".format(num_sweeps_per_beta)
            raise TypeError(error_msg)
        if num_sweeps_per_beta < 1:
            error_msg = "'num_sweeps_per_beta' should be a positive integer: value = {}".format(num_sweeps_per_beta)
            raise ValueError(error_msg)

        # handle beta_schedule et al
        if beta_schedule_type == "custom":


            if beta_schedule is None:
                error_msg = "'beta_schedule' must be provided for beta_schedule_type = 'custom': value is None"
                raise ValueError(error_msg)
            else:
                try:
                    beta_schedule = np.array(beta_schedule,dtype=float)
                except:
                    raise ValueError('beta_schedule cannot be case as a numpy array of dtype=float')
                if num_sweeps is not None and num_sweeps!=len(beta_schedule)*num_sweeps_per_beta:
                    error_msg = "'num_sweeps' should be set to None, or a value consistent with 'beta_schedule' and 'num_sweeps_per_beta' for 'beta_schedule_type' = 'custom': value = ".format(num_sweeps)
                    raise ValueError(error_msg)
                if beta_range is not None and (beta_range[0]!=beta_schedule[0] or beta_range[-1]!=beta_schedule[-1]):
                    error_msg = "'beta_range' should be set to None, or a value consistent with 'beta_schedule', for 'beta_schedule_type'='custom'."
                    raise ValueError(error_msg)
                if np.min(beta_schedule)<0:
                    error_msg = "'beta_schedule' cannot include negative values."
                    raise ValueError(error_msg)
        else:
            if beta_schedule is not None:
                error_msg = "'beta_schedule' must be set to None for 'beta_schedule_type' not equal to 'custom'"
                raise ValueError(error_msg)
            if num_sweeps is None:
                num_sweeps = 1000

            num_betas, rem = divmod(num_sweeps,num_sweeps_per_beta)

            if rem > 0 or num_betas < 0:
                error_msg = "'num_sweeps' must be a non-negative value divisible by 'num_sweeps_per_beta'."
                raise ValueError(error_msg)

            if beta_range is None:
                beta_range = _default_ising_beta_range(bqm.linear, bqm.quadratic)
            elif len(beta_range) != 2 or min(beta_range) < 0:
                error_msg = "'beta_range' should be a 2-tuple, or 2 element list of positive numbers. The latter value is the target value."
                raise ValueError(error_msg)

            if num_betas == 1:
                #One sweep in the target model
                beta_schedule = np.array([beta_range[-1]],dtype=float)
            else:
                if beta_schedule_type == "linear":
                    # interpolate a linear beta schedule
                    beta_schedule = np.linspace(*beta_range, num=num_betas)
                elif beta_schedule_type == "geometric":
                    if min(beta_range) <= 0:
                        error_msg = "'beta_range' must contain non-zero values for 'beta_schedule_type' = 'geometric'"
                        raise ValueError(error_msg)
                    # interpolate a geometric beta schedule
                    beta_schedule = np.geomspace(*beta_range, num=num_betas)
                else:
                    raise ValueError("Beta schedule type {} not implemented".format(beta_schedule_type))

        timestamp_sample = perf_counter_ns()

        # run the simulated annealing algorithm
        samples, energies = simulated_annealing(
            num_reads, ldata, irow, icol, qdata,
            num_sweeps_per_beta, beta_schedule,
            seed, initial_states_array,
            randomize_order, proposal_acceptance_criteria,
            interrupt_function)
        timestamp_postprocess = perf_counter_ns()

        info = {
            "beta_range": beta_range,
            "beta_schedule_type": beta_schedule_type
        }
        response = dimod.SampleSet.from_samples(
            (samples, variable_order),
            energy=energies+bqm.offset,  # add back in the offset
            info=info,
            vartype=dimod.SPIN
        )

        response.change_vartype(original_vartype, inplace=True)

        # Developer note: the specific keys of the timing dict are chosen to be consistent with
        #                 other samplers' timing dict.
        response.info.update(dict(timing=dict(
            preprocessing_ns=timestamp_sample - timestamp_preprocess,
            sampling_ns=timestamp_postprocess - timestamp_sample,
            # Update timing info last to capture the full postprocessing time
            postprocessing_ns=perf_counter_ns() - timestamp_postprocess,
        )))

        return response


Neal = SimulatedAnnealingSampler


def _default_ising_beta_range(h, J,
                              max_single_qubit_excitation_rate = 0.01,
                              scale_T_with_N = True):
    """Determine the starting and ending beta from h J.

    The assumed usage of annealing is optimization. Defaults are chosen
    so that at the hot temperature is fast mixing, and so that at low
    temperature the rate of excitations is small, exploiting a simple
    approximation to the ground state energy space to determine a low
    temperature bound and a worst case approximation to energy
    distribution to determine the high temperature bound.

    Args:
        h (dict):
            External field of Ising model (also called linear bias)

        J (dict):
            Couplings of Ising model (also called quadratic biases)

        max_single_qubit_excitation_rate (float, optional, default = 0.01):
            Targeted single qubit excitation rate at final temperature.
            We can set this value small to lower the probability of
            excitations at the end of the anneal, combined with a simple
            approximation to the ground-state energy structure.
            For standard schedule types (such as geometric), a trade-off
            exists between this threshold and time to solution
            (exploiting multiple restarts).

        scale_T_with_N (bool, optional, default = False):
            The expected number of excitations at finite temperature scales
            with the number of variables. For non-vanishing probability
            of the ground state it is necessary for T to scale to zero
            with N. A simplified approximation to the ground-state energy
            structure is used - for high precision problems T may not scale.

    Assume each variable in J is also in h.

    Generally one should optimize min_beta, max_beta and the shape of the schedule
    to maximize some objective such as time to solution.
    The approximations here are simplified, and can be far from optimal in some cases. Use
    custom schedules to address such cases.
    Approximations used are of complexity O(number of biases), O(1) approximations are also
    possible, as are stronger approximations, but methods are not bottlenecks in practice.
    We use the maximum bias per spin to give an upper bound on the gap, allowing the initial
    sweeps to be fast mixing.
    We use an approximation to the minimum bias per spin to give a lower bound on the minimum
    energy gap, such that for the final sweeps states are highly unlikely to excite away from
    a global (or local) minimum.
    """
    if not 0 < max_single_qubit_excitation_rate < 1:
        raise ValueError('Targeted single qubit excitations rates must be in range (0,1)')

    # Approximate worst and best cases of the [non-zero] energy signal (effective field)
    # experienced per spin as function of neighbors: bias = h_i + sum_j Jij s_j:
    sum_abs_bias_dict = defaultdict(int, {k: abs(v) for k, v in h.items()})
    if sum_abs_bias_dict:
        min_abs_bias_dict = {k: v for k, v in sum_abs_bias_dict.items() if v != 0}
    else:
        min_abs_bias_dict = {}
    #This loop is slow, but is not a bottleneck for practical implementations of simulated annealing:
    for (k1, k2), v in J.items():
        for k in [k1,k2]:
            sum_abs_bias_dict[k] += abs(v)
            if v != 0: #This is slow, but this routine is not a bottleneck:
                if k in min_abs_bias_dict:
                    min_abs_bias_dict[k] = min(abs(v),min_abs_bias_dict[k])
                else:
                    min_abs_bias_dict[k] = abs(v)

    if not min_abs_bias_dict:
        #If dictionary is empty, a Null problem is being presented. This is
        #likely a user error, or edge case test.
        #We should also consider warning for min(min_abs_bias_dict.values())==0.
        #Metropolis-Hastings is not suitable for unbiased and uncoupled
        #variables, sampling of equilibrium is possible, but only if a uniform
        #random initial condition is used (this is default, but allows changes).
        warn_msg = ('All bqm biases are zero (all energies are zero), this is '
                    'likely a value error. Temperature range is set arbitrarily '
                    'to [0.1,1]. Metropolis-Hastings update is non-ergodic.')
        warnings.warn(warn_msg)
        return([0.1,1])


    # Selecting betas based on probability of flipping a qubit
    # Hot temp: We want to be able to quickly search the entire solution space (equilibrate).
    # Scale hot_beta so that all spins are able to flip with probability at least 50%, which
    # ensures mixing across all states is fast.
    # The most unlikely flip is related to the largest energy gap that must be overcome, with
    # Metropolis updates we require:
    #   0.50 = exp(-hot_beta * max_delta_energy) - (1)
    # This is solved as hot_beta = log(2)/max_delta_energy, max_delta energy is twice the
    # effective field, we take a worst case of the effective field to be conservative.
    # Max delta energy occurs when all biases are aligned, and contribute without frustration:
    max_effective_field = max(sum_abs_bias_dict.values(), default=0)

    if max_effective_field == 0:
        hot_beta = 1
    else:
        hot_beta = np.log(2) / (2*max_effective_field)

    # Cold temp: Towards the end of the annealing schedule, we want to eliminate excitations assuming
    # an optimization application; the samples relax to a local (or ideally global) minima.
    # If we arrive at the ground state (or local minima) on the penultimate iteration, we can require
    # that we have low probability to excite on the final iteration.
    # This means that the probability to excite any spin must be bounded by 0.01 on the final sweep
    #   0.01 = sum_i exp(-cold_beta * min_delta_energy_i) - (2)
    # We can approximate min_delta_energy_i as the cost of frustrating the smallest J or h. This approximation
    # can be tight for low precision problems, but can be improved for higher precision ones. In the worst
    # case (high connectity and high precision) determining min_delta_energy_i is NP-hard (related to a
    # knapsack problem) we settle for this approximation, which yields satisfactory results in many standard
    # model types.
    # (2) is a convex optimization problem, and easy to solve with a root finder. However, for brevity we can
    # further simplify, by assuming only spins experiencing minimal gaps are excitable
    #   0.01 ~ #minimal_gaps exp(- cold_beta min_i min_delta_energy_i) - (2)
    # Where #minimal_gaps counts the number of cases i for which min_delta_energy_i = min_i(min_delta_energy_i)
    # i.e. the number of spins which are most easily excited out of the ground state.
    # The solution is cold_beta = log(#minimal_gaps/0.01) / min_delta_energy.
    # min_delta_energy is twice the smallest (non-zero) effective field, we approximate this per spin as the
    # smallest associated bias term (h or J).
    if len(min_abs_bias_dict)==0:
        #Trivial problem
        cold_beta = hot_beta
    else:
        values_array = np.array(list(min_abs_bias_dict.values()),dtype=float)
        min_effective_field = np.min(values_array)
        if scale_T_with_N:
            #This is effective assuming finite precision, for pathological
            #or real valued problems this approximation may perform poorly
            number_min_gaps = np.sum(min_effective_field == values_array)
        else:
            number_min_gaps = 1
        cold_beta = np.log(number_min_gaps/max_single_qubit_excitation_rate) / (2*min_effective_field)

    return [hot_beta, cold_beta]

def default_beta_range(bqm):
    ising = bqm.spin
    return _default_ising_beta_range(ising.linear, ising.quadratic)



# Copyright 2022 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS F ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from collections import OrderedDict, namedtuple

import networkx as nx

Edge = namedtuple('Edge', ['head', 'tail', 'key'])
"""Represents a (directed) edge in a Multigraph"""


def rotation_from_coordinates(G: nx.MultiGraph, pos: dict) -> dict:
    """Compute the rotation system for a planar G from the node positions.

    Args:
        G (:obj:`networkx.MultiGraph`):
            A planar MultiGraph.

        pos (callable/dict):
            The position for each node.

    Returns:
        dict: Dictionary of nodes with a rotation dict as the value.

    """
    if not isinstance(G, nx.MultiGraph):
        raise TypeError("expected G to be a MultiGraph")

    rotation = {}
    for u in G.nodes:
        x0, y0 = pos[u]

        def angle(edge):
            if len(edge) == 3:  # generic for Graph or MultiGraph
                _, v, _ = edge
            else:
                _, v = edge
            x, y = pos[v]
            return math.atan2(y - y0, x - x0)

        if isinstance(G, nx.MultiGraph):
            circle = sorted(G.edges(u, keys=True), key=angle)
        else:
            circle = sorted(G.edges(u), key=angle)

        circle = [Edge(*edge) for edge in circle]

        rotation[u] = OrderedDict((circle[i - 1], edge) for i, edge in enumerate(circle))

    return rotation


def plane_triangulate(G: nx.MultiGraph):
    """Add edges to planar graph G to make it plane triangulated.

    An embedded graph is plane triangulated iff it is biconnected and
    each of its faces is a triangle.

    Args:
        G (:obj:`nx.MultiGraph`):
            A planar graph. Must have a rotation system. Note that edges are added in-place.

    """

    if len(G) < 3:
        raise ValueError("only defined for graphs with 3 or more nodes")

    rotation_system = {v: G.nodes[v]['rotation'] for v in G}

    # following the notation from the paper
    for i in G.nodes:
        for edge in list(G.edges(i, keys=True)):
            i, j, _ = ij = Edge(*edge)

            j, k, _ = jk = rotation_system[j][(j, i, ij.key)]
            k, l, _ = kl = rotation_system[k][(k, j, jk.key)]

            assert ij.tail == jk.head

            while l != i:
                m, n, _ = rotation_system[l][(l, k, kl.key)]
                if (m, n) == (l, j):
                    break

                if i == k:
                    # avoid self-loop
                    i, j, _ = ij = jk
                    j, k, _ = jk = kl
                    k, l, _ = kl = rotation_system[k][(k, j, jk.key)]

                    assert ij.tail == jk.head

                _insert_chord(ij, jk, G, rotation_system)

                i, j, _ = ij = kl
                j, k, _ = jk = rotation_system[j][(j, i, ij.key)]
                k, l, _ = kl = rotation_system[k][(k, j, jk.key)]

                assert ij.tail == jk.head

    assert is_plane_triangulated(G), "Something went wrong, G is not plane triangulated"

    return


def is_plane_triangulated(G: nx.MultiGraph) -> bool:
    """
    Args:
        G: The graph to be tested; note that this function expects G to have the rotation system as node attributes
    Returns:
        True iff G is bi-connected and each face is triangular
    """
    if not nx.is_biconnected(G):
        return False

    # x
    # |\
    # | \
    # y--z
    for x in G.nodes:
        for xz in G.edges(x, keys=True):

            _ = x, y, xykey = G.nodes[x]['rotation'][xz]
            _ = y, z, yzkey = G.nodes[y]['rotation'][(y, x, xykey)]
            _ = z, x, xzkey = G.nodes[z]['rotation'][(z, y, yzkey)]

            if xz != (x, z, xzkey):
                return False
    return True


def odd_in_degree_orientation(H: nx.MultiGraph) -> dict:
    """
    Args:
        H:
    Returns:
    """
    G = H.copy()

    orientation = set()

    for (u, v) in reversed(list(nx.dfs_edges(H))):
        uv = u, v, uvkey = u, v, min(G[u][v])

        uv_odd = True  # for now assume that we'll mark uv as odd

        # TODO: Should "keys=True" be "data=True"? "keys" seems to be unsupported?
        for vw in G.edges(v, keys=True):
            v, w, vwkey = vw

            if vw == (v, u, uvkey):
                # we'll do this one last
                pass
            elif (w, v, vwkey) in orientation:
                # we've already done this one and it's heading in, so we need to toggle the edge
                # we walked in on
                uv_odd = not uv_odd
                continue
            else:
                # ok, it's not the edge we came in on, and it's not already been marked or it's already
                # going out, so let's just set it going out so it doesn't change our degree
                orientation.add(vw)

        if uv_odd:
            # we want uv oriented towards v
            orientation.add(uv)
        else:
            orientation.add((v, u, uvkey))

    return {(u, v, key): v for (u, v, key) in orientation}


def expanded_dual(G: nx.MultiGraph) -> nx.Graph:
    """
    Args:
        G: should be multigraph, triangulated, oriented, edges indexed
    """

    dual = nx.Graph()

    # first we add the edges of the dual that cross the edges of G
    # for an edge (u, v, key) oriented towards v, we adopt the convention
    # that the right-hand node is labelled (u, v, key) and the left-hand
    # node is (v, u, key).
    for edge in G.edges(keys=True):
        u = edge
        v = (edge[1], edge[0], edge[2])
        dual.add_edge(u, v, weight=G.edges[edge].get('weight', 0.0))

    # next we add the edges within each triangular face
    for n in G.nodes:
        # iterate through the edges around n
        for left in G.edges(n, keys=True):
            u, v, _ = left = Edge(*left)
            assert u == n
            s, t, _ = right = Edge(*G.nodes[u]['rotation'][left])
            assert s == u

            # we want to connect the node left (from n) or right
            dual.add_edge(tuple(left), (t, s, right.key), weight=0.0)

    return dual


def _inverse_rotation_system(rotation_system: dict, v: str, edge: Edge) -> Edge:
    for e1, e2 in rotation_system[v].items():
        if e2 == edge:
            return e1

    raise RuntimeError


def _insert_chord(ij: Edge, jk: Edge, G: nx.MultiGraph, rotation_system: dict):
    """Insert a chord between i and k."""
    assert ij.tail == jk.head
    i, j, _ = ij
    j, k, _ = jk

    # because G is a Multigraph, G.add_edge returns the key
    ik = Edge(i, k, G.add_edge(i, k))

    rotation_system[k][(k, i, ik.key)] = rotation_system[k][(k, j, jk.key)]
    rotation_system[k][(k, j, jk.key)] = Edge(k, i, ik.key)

    rotation_system[i][_inverse_rotation_system(rotation_system, i, ij)] = ik
    rotation_system[i][ik] = ij



# distutils: language = c++
# distutils: include_dirs = greedy/src/
# distutils: include_dirs = dwave/samplers/greedy/src/
# distutils: sources = dwave/samplers/greedy/src/descent.cpp

# Copyright 2019 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from libcpp cimport bool
from libcpp.vector cimport vector

import numpy as np
cimport numpy as np

cimport decl


def steepest_gradient_descent(num_samples,
                              linear_biases,
                              coupler_starts, coupler_ends, coupler_weights,
                              np.ndarray[np.int8_t, ndim=2, mode="c"] states_numpy,
                              large_sparse_opt=False):

    """Wraps `steepest_gradient_descent` from `descent.cpp`. Accepts
    an Ising problem defined on a general graph and returns samples
    using steepest gradient descent seeded with initial states from
    `states_numpy`.

    Parameters
    ----------
    num_samples : int
        Number of samples to get from the sampler.

    linear_biases : list(float)
        The linear biases or field values for the problem.

    coupler_starts : list(int)
        A list of the start variable of each coupler. For a problem
        with the couplers (0, 1), (1, 2), and (3, 1), `coupler_starts`
        should be [0, 1, 3].

    coupler_ends : list(int)
        A list of the end variable of each coupler. For a problem
        with the couplers (0, 1), (1, 2), and (3, 1), `coupler_ends`
        should be [1, 2, 1].

    coupler_weights : list(float)
        A list of the J values or weight on each coupler, in the same
        order as `coupler_starts` and `coupler_ends`.

    states_numpy : np.ndarray[np.int8_t, ndim=2, mode="c"], values in (-1, 1)
        The initial seeded states of the gradient descent runs. Should be of
        a contiguous numpy.ndarray of shape (num_samples, num_variables).

    large_sparse_opt : bool
        When set to True, large-and-sparse problem graph optimizations are used.

    Returns
    -------
    samples : numpy.ndarray
        A 2D numpy array where each row is a sample.

    energies: numpy.ndarray
        Sample energies.

    num_steps: numpy.ndarray
        Number of downhill steps per sample.
    """
    num_vars = len(linear_biases)

    # short-circuit null edge cases
    if num_samples == 0 or num_vars == 0:
        states = np.empty((num_samples, num_vars), dtype=np.int8)
        return (states,
                np.zeros(num_samples, dtype=np.double),
                np.zeros(num_samples, dtype=np.uint32))

    # allocate ndarray for energies
    energies_numpy = np.empty(num_samples, dtype=np.float64)
    cdef double[:] energies = energies_numpy

    # allocate ndarray for steps
    num_steps_numpy = np.empty(num_samples, dtype=np.uint32)
    cdef unsigned[:] num_steps = num_steps_numpy

    # explicitly convert all Python types to C while we have the GIL
    cdef np.int8_t* _states = &states_numpy[0, 0]
    cdef double* _energies = &energies[0]
    cdef unsigned* _num_steps = &num_steps[0]
    cdef int _num_samples = num_samples
    cdef bool _large_sparse_opt = large_sparse_opt
    cdef vector[double] _linear_biases = linear_biases

    # TODO: in dimod 0.10+, coupler indices default to int64, but we downcast
    # them still to int. We can update this when we upgrade to a more efficient
    # bqm data interface. Until then, beware of the 2B num_variables limit.
    cdef vector[int] _coupler_starts = coupler_starts
    cdef vector[int] _coupler_ends = coupler_ends
    cdef vector[double] _coupler_weights = coupler_weights

    with nogil:
        decl.steepest_gradient_descent(
            _states, _energies, _num_steps, _num_samples,
            _linear_biases, _coupler_starts, _coupler_ends, _coupler_weights,
            _large_sparse_opt)

    return states_numpy, energies_numpy, num_steps_numpy
